import requests


def get_download_links(
    session: requests.Session,
    model: str,
    branch: str,
    text_only: bool = False,
    specific_file: str = None,
) -> (list, list, bool, bool):
    """
    Gets download links for the specified model and branch from Hugging Face.

    Args:
        session (requests.Session): The session object to handle HTTP requests.
        model (str): The model name.
        branch (str): The branch name.
        text_only (bool, optional): If True, only text files will be downloaded. Defaults to False.
        specific_file (str, optional): Specific file to download. Defaults to None.

    Returns:
        tuple: A tuple containing the list of download links, sha256 checksums,
               a boolean indicating if it's a LORA model, and a boolean for the LLAMAcpp type.
    """
    import json
    import re
    import base64

    base_url = "https://huggingface.co"
    page = f"/api/models/{model}/tree/{branch}"
    cursor = b""

    links = []
    sha256 = []
    classifications = []
    has_pytorch = False
    has_pt = False
    has_gguf = False
    has_safetensors = False
    is_lora = False

    while True:
        url = f"{base_url}{page}" + (f"?cursor={cursor.decode()}" if cursor else "")
        response = session.get(url, timeout=10)
        response.raise_for_status()
        content = response.content

        data = json.loads(content)
        if len(data) == 0:
            break

        for item in data:
            fname = item["path"]
            if specific_file not in [None, ""] and fname != specific_file:
                continue

            if not is_lora and fname.endswith(
                ("adapter_config.json", "adapter_model.bin")
            ):
                is_lora = True

            is_pytorch = re.match(r"(pytorch|adapter|gptq)_model.*\.bin", fname)
            is_safetensors = re.match(r".*\.safetensors", fname)
            is_pt = re.match(r".*\.pt", fname)
            is_gguf = re.match(r".*\.gguf", fname)
            is_tiktoken = re.match(r".*\.tiktoken", fname)
            is_tokenizer = (
                re.match(r"(tokenizer|ice|spiece).*\.model", fname) or is_tiktoken
            )
            is_text = re.match(r".*\.(txt|json|py|md)", fname) or is_tokenizer

            if any((is_pytorch, is_safetensors, is_pt, is_gguf, is_tokenizer, is_text)):
                if "lfs" in item:
                    sha256.append([fname, item["lfs"]["oid"]])

                if is_text:
                    links.append(
                        f"https://huggingface.co/{model}/resolve/{branch}/{fname}"
                    )
                    classifications.append("text")
                    continue

                if not text_only:
                    links.append(
                        f"https://huggingface.co/{model}/resolve/{branch}/{fname}"
                    )
                    if is_safetensors:
                        has_safetensors = True
                        classifications.append("safetensors")
                    elif is_pytorch:
                        has_pytorch = True
                        classifications.append("pytorch")
                    elif is_pt:
                        has_pt = True
                        classifications.append("pt")
                    elif is_gguf:
                        has_gguf = True
                        classifications.append("gguf")

        cursor = (
            base64.b64encode(f'{{"file_name":"{data[-1]["path"]}"}}'.encode()) + b":50"
        )
        cursor = base64.b64encode(cursor)
        cursor = cursor.replace(b"=", b"%3D")

    def remove_pytorch_pt_links(links, classifications):
        new_links = []
        new_classifications = []
        for link, classification in zip(links, classifications):
            if classification not in ["pytorch", "pt"]:
                new_links.append(link)
                new_classifications.append(classification)
        return new_links, new_classifications

    def process_gguf_links(links, classifications, specific_file):
        if specific_file is None:
            has_q4km = any("q4_k_m" in link.lower() for link in links)
            if has_q4km:
                return [link for link in links if "q4_k_m" in link.lower()], [
                    "gguf" for link in links if "q4_k_m" in link.lower()
                ]
            else:
                return [link for link in links if not link.lower().endswith(".gguf")], [
                    classification
                    for link, classification in zip(links, classifications)
                    if not link.lower().endswith(".gguf")
                ]
        return links, classifications

    # Additional processing based on file types
    if (has_pytorch or has_pt) and has_safetensors:
        links, classifications = remove_pytorch_pt_links(links, classifications)

    if has_gguf:
        links, classifications = process_gguf_links(
            links, classifications, specific_file
        )

    is_llamacpp = has_gguf and specific_file is not None
    return links, sha256, is_lora, is_llamacpp



def download(
    *,
    model,
    branch="main",
    threads=4,
    text_only=False,
    specific_file=None,
    output=None,
    clean=False,
    check=True,
    max_retries=5,
):
    """
    Downloads models from Hugging Face to models/username_modelname.

    Example:
    python download-model.py facebook/opt-1.3b

    """

    import base64
    import json
    import os
    import re
    import sys
    from pathlib import Path
    import hashlib
    import requests
    import tqdm
    from requests.adapters import HTTPAdapter

    base = "https://huggingface.co"

    class _ModelDownloader:
        def __init__(self, max_retries=5):
            self.session = requests.Session()
            if max_retries:
                self.session.mount(
                    "https://cdn-lfs.huggingface.co",
                    HTTPAdapter(max_retries=max_retries),
                )
                self.session.mount(
                    "https://huggingface.co", HTTPAdapter(max_retries=max_retries)
                )
            if os.getenv("HF_USER") is not None and os.getenv("HF_PASS") is not None:
                self.session.auth = (os.getenv("HF_USER"), os.getenv("HF_PASS"))
            if os.getenv("HF_TOKEN") is not None:
                self.session.headers = {
                    "authorization": f'Bearer {os.getenv("HF_TOKEN")}'
                }

        def check_model_files(self, model, branch, links, sha256, output_folder):
            # Validate the checksums
            validated = True
            for i in range(len(sha256)):
                fpath = output_folder / sha256[i][0]

                if not fpath.exists():
                    print(f"The following file is missing: {fpath}")
                    validated = False
                    continue

                with open(output_folder / sha256[i][0], "rb") as f:
                    file_hash = hashlib.file_digest(f, "sha256").hexdigest()
                    if file_hash != sha256[i][1]:
                        print(f"Checksum failed: {sha256[i][0]}  {sha256[i][1]}")
                        validated = False
                    else:
                        print(f"Checksum validated: {sha256[i][0]}  {sha256[i][1]}")

            if validated:
                print("[+] Validated checksums of all model files!")
            else:
                print(
                    "[-] Invalid checksums. Rerun download-model.py with the --clean flag."
                )

    if model is None:
        print(
            "Error: Please specify the model you'd like to download (e.g. 'download(model='facebook/opt-1.3b')."
        )
        sys.exit()

    downloader = _ModelDownloader(max_retries=max_retries)

    # Clean up the model/branch names
    import A_GIS.Data.Llm._helpers

    model, branch = A_GIS.Data.Llm._helpers.sanitize_model_and_branch_names(
        model, branch
    )

    # Get the download links from Hugging Face
    (
        links,
        sha256,
        is_lora,
        is_llamacpp,
    ) = A_GIS.Data.Llm._helpers.get_download_links(
        downloader.session,
        model,
        branch,
        text_only=text_only,
        specific_file=specific_file,
    )

    # Get the output folder
    output_folder = A_GIS.Data.Llm._helpers.get_output_folder(
        model, branch, is_lora, is_llamacpp=is_llamacpp, base_folder=output
    )

    if check:
        # Check previously downloaded files
        downloader.check_model_files(model, branch, links, sha256, output_folder)
    else:
        # Download files
        A_GIS.Data.Llm._helpers.download_model_files(
            downloader.session,
            model,
            branch,
            links,
            sha256,
            output_folder,
            specific_file=specific_file,
            threads=threads,
            is_llamacpp=is_llamacpp,
        )

    return output_folder


if __name__ == "__main__":
    import A_GIS.Data.Llm._helpers.parse_download_cli_args

    args = A_GIS.Data.Llm._helpers.parse_download_cli_args()

    download(
        model=args.MODEL,
        branch=args.branch,
        threads=args.threads,
        text_only=args.text_only,
        specific_file=args.specific_file,
        output=args.output,
        clean=args.clean,
        check=args.check,
        max_retries=args.max_retries,
    )



def parse_download_cli_args():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("MODEL", type=str, default=None, nargs="?")
    parser.add_argument(
        "--branch",
        type=str,
        default="main",
        help="Name of the Git branch to download from.",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=4,
        help="Number of files to download simultaneously.",
    )
    parser.add_argument(
        "--text-only", action="store_true", help="Only download text files (txt/json)."
    )
    parser.add_argument(
        "--specific-file",
        type=str,
        default=None,
        help="Name of the specific file to download (if not provided, downloads all).",
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="The folder where the model should be saved.",
    )
    parser.add_argument(
        "--clean", action="store_true", help="Does not resume the previous download."
    )
    parser.add_argument(
        "--check", action="store_true", help="Validates the checksums of model files."
    )
    parser.add_argument(
        "--max-retries",
        type=int,
        default=5,
        help="Max retries count when get error in download time.",
    )
    return parser.parse_args()



import PIL.Image
import typing


def get(*, image: PIL.Image.Image) -> typing.Dict[str, str]:
    """
    Retrieve the metadata of an in-memory image object as a dictionary.

    This function extracts the metadata from a given PIL.Image.Image object.
    The metadata is stored as JSON-serialized strings within the image object.
    This function deserializes these strings back into their original dictionary format.

    Args:
        image (PIL.Image.Image): The image object from which to retrieve metadata.

    Returns:
        typing.Dict[str, str]: A dictionary containing the deserialized metadata
        of the image. Each key-value pair represents a metadata field and its value.

    Examples:
        >>> import A_GIS.Image
        >>> image = A_GIS.Image.new(size=(100,100), metadata={'author':'me'})
        >>> metadata = A_GIS.Image.Metadata.get(image=image)
        >>> metadata['author']
        'me'

    Raises:
        ValueError: If the image's metadata is not in the expected format.
    """
    import json
    import sys

    metadata = {}
    if hasattr(image, "info"):
        for key, value in image.info.items():
            try:
                # Deserialize the metadata value from JSON
                metadata[key] = json.loads(value)
            except json.JSONDecodeError:
                raise ValueError(
                    f"Metadata for key '{key}' is not in valid JSON format."
                )

    return metadata



def modify(code: str, docstring: str):
    """
    Add or replace the docstring of the first class or function definition in the code.

    Parameters:
    code (str): The original Python code.
    new_docstring (str): The new docstring to be added or replaced.

    Returns:
    str: The modified Python code with the new docstring.
    """
    import re

    # Pattern to match class or function definitions
    pattern = re.compile(
        r'^(class\s+\w+|def\s+\w+)\s*\(.*?\)[^:]*:\s*(""".*?""")?(\s*$)+',
        re.DOTALL | re.MULTILINE,
    )

    # Format the new docstring
    if docstring == None:
        formatted_docstring = ""
    else:
        formatted_docstring = f'"""{docstring.strip()}\n"""\n'

    # Function to replace or add the docstring
    def _replace_match(match):
        existing_docstring = match.group(2)
        if existing_docstring:
            # Replace existing docstring
            return re.sub(
                r'""".*?"""(\s*$)+',
                formatted_docstring,
                match.group(0),
                count=1,
                flags=re.DOTALL | re.MULTILINE,
            )
        else:
            # Add new docstring
            if docstring == None:
                return match.group(0)
            else:
                return (
                    match.group(0)
                    + "\n"
                    + re.sub("^", "    ", formatted_docstring, flags=re.MULTILINE)
                )

    # Replace or add the docstring in the first class or function definition
    new_code, num_replacements = pattern.subn(_replace_match, code, count=1)

    return new_code



def apply_patch(*, text: str, patch: str):
    """
    Apply a unified diff to a string.

    Parameters:
    original_text (str): The original text to which the diff will be applied.
    unified_diff (str): The unified diff as a string.

    Returns:
    str: The modified text after applying the diff.
    """
    from diff_match_patch import diff_match_patch

    dmp = diff_match_patch()

    # Convert the unified diff string to a list of Patch objects
    patches = dmp.patch_fromText(patch)

    # Apply the patches to the original text
    patched_text, _ = dmp.patch_apply(patches, text)

    return patched_text



def open(*, path: str):
    """
    Open an image file from either a local file path or a URL.

    This function uses the PIL library to open a local image file, and the requests library
    to fetch and open an image from a URL.

    Args:
        source (str): The path or URL of the image to be opened.

    Returns:
        PIL.Image.Image: An image object opened from the given path or URL.

    Example:
        >>> import A_GIS.Image
        >>> image = A_GIS.Image.open(path='https://picsum.photos/200')
    """
    from PIL import Image
    from io import BytesIO
    import requests

    # Check if the source is a URL
    if path.startswith("http://") or path.startswith("https://"):
        # Fetch the image from the URL
        response = requests.get(path)
        response.raise_for_status()  # This will raise an exception for HTTP errors

        # Open and return the image from the response content
        return Image.open(BytesIO(response.content))
    else:
        # Open and return the image from a local file
        return Image.open(path)



import PIL.Image
import typing


def new(
    *,
    mode: str = "RGBA",
    size: typing.Tuple[int, int] = (0, 0),
    metadata: typing.Optional[typing.Dict[str, str]] = None,
) -> PIL.Image.Image:
    """
    Create and return a new image with the specified mode, size, and optional metadata.

    This function utilizes the Python Imaging Library (PIL) to create a new image object.
    The mode specifies the color format, such as 'RGB' or 'RGBA'. The size is defined by a tuple
    (width, height). If metadata is provided, it is added to the image as a dictionary of JSON strings.

    Args:
        mode: The color mode to use for the new image. Defaults to 'RGBA'.
        size: A tuple (width, height) specifying the dimensions of the new image. Defaults to (0, 0).
        metadata: Optional dictionary containing metadata for the image. Each key-value pair is added
                  as a JSON-formatted string. Defaults to None.

    Returns:
        A new PIL.Image.Image object with the specified properties.

    Examples:
        >>> new_image = new(mode='RGB', size=(200, 200), metadata={'Author': 'John Doe', 'Description': 'Sample Image'})
        >>> empty_image = new()

    Raises:
        ImportError: If PIL modules are not available.
    """
    import PIL.Image

    # Create a new image object with the specified mode and size
    image = PIL.Image.new(mode, size)

    # Add metadata to the image if provided which requires writing to disk with PIL.
    if metadata:
        import A_GIS.File.Directory
        import pathlib
        import json
        import PIL.PngImagePlugin

        pnginfo = PIL.PngImagePlugin.PngInfo()
        for key, value in metadata.items():
            pnginfo.add_text(key, json.dumps(value))
        tempdir = A_GIS.File.Directory.make(scoped_delete=True)
        path = pathlib.Path(tempdir.path) / "new.png"
        image.save(path, "PNG", pnginfo=pnginfo)
        image = PIL.Image.open(path)

    return image



def distill(*, code: str) -> str:
    """Purify the given Python code by removing docstrings, blanking multiline
    string literals, and inherently removing comments.

    This function parses the provided Python code into an abstract syntax tree (AST),
    and then walks through the AST. It identifies and blanks out all docstrings and
    multiline string literals. Additionally, due to the nature of AST parsing in Python,
    comments (which start with '#') are not included in the AST and hence are not present
    in the unparsed code, effectively removing them from the output.

    It's important to note that this removal of comments is a byproduct of how the Python
    parser and the AST handle comments, rather than an explicit action by this function.

    Args:
        code (str): A string representing the Python code to be purified.

    Returns:
        str: A string representing the purified Python code, with docstrings, multiline string
             literals, and comments removed.

    Examples:
        >>> ds=('"'*3) + 'I am a docstring!' + ('"'*3)
        >>> test_code = f'''
        ... def example_function(param1, param2):
        ...     {ds}
        ...     # This is a comment.
        ...     return (param1, param2)
        ... '''
        >>> print(distill(code=test_code))
        def example_function(param1, param2):
            return (param1, param2)
    """

    # Parse the code into an abstract syntax tree (AST).
    import ast

    parsed = ast.parse(code)

    # Walk through all nodes in the AST.
    for node in ast.walk(parsed):
        # Identify docstrings and multiline string literals, and replace them
        # with empty strings.
        if isinstance(node, ast.Expr) and isinstance(node.value, ast.Str):
            node.value = ast.Constant(value="")

    # Remove any blank lines or blank docstrings.
    import re

    distilled_code = re.sub(
        r'^\s*""""""\s*$\n', "", ast.unparse(parsed), flags=re.MULTILINE
    )
    return re.sub(r"\n\s*\n", "\n", distilled_code)



def update(*, tree: dict):
    """Update the package files throughout the files listed in the tree to have the
    correct format and content.
    """
    import A_GIS.File.write

    imports = None
    if "_type" in tree:
        if tree["_type"] == "package":
            imports = {"Functions": [], "Classes": [], "Packages": []}
    for name in tree:
        if name in set(["_type", "_file"]):
            continue
        if imports:
            if tree[name]["_type"] == "package":
                imports["Packages"].append(f"from . import {name}")
            elif tree[name]["_type"] == "function":
                imports["Functions"].append(f"from .{name} import {name}")
            elif tree[name]["_type"] == "class":
                imports["Classes"].append(f"from .{name} import {name}")

        update(tree=tree[name])

    if imports:
        code = ""
        for k, v in imports.items():
            if len(v) > 0:
                code += "# {}\n{}\n".format(k, "\n".join(sorted(v)))
        A_GIS.File.write(content=code, file=tree["_file"])



# Functions
from .download_model_files import download_model_files
from .get_download_links import get_download_links
from .get_output_folder import get_output_folder
from .parse_download_cli_args import parse_download_cli_args
from .sanitize_model_and_branch_names import sanitize_model_and_branch_names
from .start_download_threads import start_download_threads



import os
import pytest
import A_GIS.File.Directory


def test_tempdir_creation():
    """Test if the TempDir creates a directory."""
    with A_GIS.File.Directory.make() as temp_dir:
        assert os.path.isdir(temp_dir.path), "Directory should be created"


def test_tempdir_custom_path():
    """Test TempDir with a custom path."""
    custom_path = "my_temp_dir"
    with A_GIS.File.Directory.make(path=custom_path) as temp_dir:
        assert (
            temp_dir.path == custom_path
        ), "Directory path should match the custom path"
        assert os.path.isdir(custom_path), "Custom directory should be created"
    # Clean up
    if os.path.exists(custom_path):
        os.rmdir(custom_path)


def test_tempdir_scoped_delete():
    """Test if the TempDir deletes itself when scoped_delete is True."""
    with A_GIS.File.Directory.make(scoped_delete=True) as temp_dir:
        path = temp_dir.path
    assert not os.path.exists(
        path
    ), "Directory should be deleted after exiting the context block"


def test_tempdir_no_delete():
    """Test if the TempDir does not delete itself when scoped_delete is False."""
    with A_GIS.File.Directory.make(scoped_delete=False) as temp_dir:
        path = temp_dir.path
    assert os.path.exists(
        path
    ), "Directory should not be deleted after exiting the context block"
    # Clean up
    os.rmdir(path)


def test_tempdir_destructor():
    """Test the destructor of TempDir for self-deletion."""
    temp_dir = A_GIS.File.Directory.make(scoped_delete=True)
    path = temp_dir.path
    del temp_dir
    assert not os.path.exists(
        path
    ), "Directory should be deleted when the object is destructed"



def from_samples(*, realizations):
    """Calculate correlation coefficient from samples."""
    import numpy
    import A_GIS.Math.Statistics.CorrelationMatrix.from_numpy

    matrix = numpy.corrcoef(realizations, rowvar=False)
    return A_GIS.Math.Statistics.CorrelationMatrix.from_numpy(matrix=matrix)



def make(*, path: str = None, scoped_delete: bool = False):
    """
    Creates a directory object that may delete itself when it goes out of scope.

    This function returns an instance of the nested class `TempDir`. This class
    provides functionality to create a temporary directory that can be set to
    self-delete when the object is no longer in use, based on the `scoped_delete` flag.

    Args:
        path (str, optional): The path where the directory should be created. If None,
                              a temporary directory is created using the `tempfile` module.
        scoped_delete (bool, optional): If True, the created directory will be deleted
                                        when the TempDir object is destroyed or when
                                        exiting a context manager block.

    Returns:
        TempDir: An instance of the TempDir class representing the created directory.

    Examples:
        >>> import A_GIS.File.Directory.make
        >>> with A_GIS.File.Directory.make(scoped_delete=True) as temp_dir:
        ...     pass
    """
    import os
    import tempfile
    import shutil
    class _TempDir:
        def __init__(self, path: str = None, scoped_delete: bool = False):
            """Initializes the TempDir object."""
            self.scoped_delete = scoped_delete
            self.path = path if path else tempfile.mkdtemp()

            if path and not os.path.exists(self.path):
                os.makedirs(self.path)

        def __enter__(self):
            """Called when entering the 'with' block."""
            return self

        def __exit__(self, exc_type, exc_val, exc_tb) -> None:
            """Called when exiting the 'with' block."""
            if self.scoped_delete:
                self._delete_dir()

        def __del__(self) -> None:
            """Called when the object is about to be destroyed."""
            if self.scoped_delete:
                self._delete_dir()

        def _delete_dir(self) -> None:
            """Helper method to delete the directory."""
            if os.path.exists(self.path):
                shutil.rmtree(self.path)

    return _TempDir(path, scoped_delete)



import requests
from pathlib import Path


def download(
    session: requests.Session,
    url: str,
    output_folder: Path,
    start_from_scratch: bool = False,
):
    """
    Downloads a single file from a given URL.

    Args:
        session (requests.Session): The session object to handle HTTP requests.
        url (str): The URL from which to download the file.
        output_folder (Path): The folder where the file will be saved.
        start_from_scratch (bool, optional): If True, any existing file will be overwritten. Defaults to False.
    """
    import tqdm
    import os

    filename = Path(url.rsplit("/", 1)[1])
    output_path = output_folder / filename
    headers = {}
    mode = "wb"

    if output_path.exists() and not start_from_scratch:
        # Resume the download from where it left off
        r = session.get(url, stream=True, timeout=10)
        total_size = int(r.headers.get("content-length", 0))
        if output_path.stat().st_size >= total_size:
            return

        headers = {"Range": f"bytes={output_path.stat().st_size}-"}
        mode = "ab"

    with session.get(url, stream=True, headers=headers, timeout=10) as r:
        r.raise_for_status()
        total_size = int(r.headers.get("content-length", 0))
        block_size = 1024 * 1024  # 1MB

        tqdm_kwargs = {
            "total": total_size,
            "unit": "iB",
            "unit_scale": True,
            "bar_format": "{l_bar}{bar}| {n_fmt:6}/{total_fmt:6} {rate_fmt:6}",
        }

        if "COLAB_GPU" in os.environ:
            tqdm_kwargs.update({"position": 0, "leave": True})

        with open(output_path, mode) as f:
            with tqdm.tqdm(**tqdm_kwargs) as t:
                for data in r.iter_content(block_size):
                    t.update(len(data))
                    f.write(data)



from pathlib import Path


def get_output_folder(
    model: str,
    branch: str,
    is_lora: bool,
    is_llamacpp: bool = False,
    base_folder: str = None,
) -> Path:
    """
    Determines the output folder for the downloaded model.

    Args:
        model (str): The model name.
        branch (str): The branch name.
        is_lora (bool): Indicates if the model is a LORA model.
        is_llamacpp (bool, optional): Indicates if the model is of type GGUF. Defaults to False.
        base_folder (str, optional): Base folder for the model storage. Defaults to None.

    Returns:
        Path: The calculated output folder as a Path object.
    """
    if base_folder is None:
        base_folder = "models" if not is_lora else "loras"

    if is_llamacpp:
        return Path(base_folder)

    output_folder = f"{'_'.join(model.split('/')[-2:])}"
    if branch != "main":
        output_folder += f"_{branch}"

    return Path(base_folder) / output_folder



def print(*, tree, indent=0):
    import A_GIS.Code.Tree.print

    markdown = ""
    indent_str = "    " * indent  # 4 spaces for each indentation level

    for key, value in tree.items():
        if key == "_type" or key == "_file":
            continue
        markdown += f"{indent_str}- {value['_type']} {key}\n"

        # Recursively process nested items
        if isinstance(value, dict) and value:
            markdown += A_GIS.Code.Tree.print(tree=value, indent=indent + 1)

    return markdown



import pathlib
import typing


def read(*, source: typing.Union[str, pathlib.Path]) -> typing.Dict[str, typing.Any]:
    """
    Reads JSON data from a string or file and returns its contents as a dictionary.

    This function accepts either a JSON string or a file path encapsulated in a `pathlib.Path` object.
    It attempts to parse the input into a dictionary. If the input is a `pathlib.Path` object, it reads
    the file located at that path. If the input is a string, it is treated as JSON data.

    Args:
        source: A string representing JSON data or a `pathlib.Path` object representing the file path
                to a JSON file.

    Returns:
        A dictionary representing the parsed JSON.

    Raises:
        json.JSONDecodeError: If the input string is not a valid JSON.
        FileNotFoundError: If the specified file in `pathlib.Path` does not exist.
        IOError: For issues related to file reading.
        TypeError: If the input is neither a string nor a `pathlib.Path` object.

    Example:
        data = read(source=pathlib.Path('path/to/your/file.json'))
        # or
        data = read(source='{"key": "value"}')
    """
    import json

    if isinstance(source, pathlib.Path):
        with open(source, "r") as file:
            return json.load(file)
    else:
        return json.loads(source)



def get(*, code: str, clean: bool = True) -> str | None:
    """
    Extracts the docstring from a string of Python code.

    This function parses the provided Python code into an abstract syntax tree (AST),
    and then retrieves the docstring of the first module-level entity (usually a
    function, class, or the module itself).

    Parameters:
    code: A string representing the Python code.

    Returns:
    The docstring found in the code, if any; otherwise, returns None.
    """
    import ast

    # Parse the code into an abstract syntax tree (AST)
    tree = ast.parse(code)

    # Get the docstring of the first module-level entity
    for x in tree.body:
        try:
            return ast.get_docstring(x, clean=clean)
        except:
            pass

    return None



import re


def get(
    *, code: str, ignore_class=re.compile("^_"), ignore_function=re.compile("^_")
) -> dict:
    """Get a code tree from a code string."""
    import ast

    class _HierarchyVisitor(ast.NodeVisitor):
        def __init__(self):
            self.hierarchy = {}
            self.current_scope = self.hierarchy

        def visit_ClassDef(self, node):
            if ignore_class.match(node.name):
                return
            class_name = node.name
            previous_scope = self.current_scope
            self.current_scope = self.current_scope.setdefault(
                class_name, {"_type": "class"}
            )
            for child_node in node.body:
                self.visit(child_node)
            self.current_scope = previous_scope

        def visit_FunctionDef(self, node):
            if ignore_function.match(node.name):
                return
            function_name = node.name
            previous_scope = self.current_scope
            self.current_scope[function_name] = {"_type": "function"}
            self.current_scope = self.current_scope[function_name]
            for child_node in node.body:
                self.visit(child_node)
            self.current_scope = previous_scope

    nodes = ast.parse(code)
    hierarchy_visitor = _HierarchyVisitor()
    hierarchy_visitor.visit(nodes)

    return hierarchy_visitor.hierarchy



import typing


def new(
    *, size: int = 1, values: typing.Optional[typing.List[float]] = []
):
    """
    Return a correlation matrix data class from the upper triangular values.

    This function initializes a `_CorrMatrix` data class with a specified size and set of values.
    The values are expected to represent the upper triangular part of the matrix.

    Args:
        size (int): The size of the correlation matrix. Defaults to 1.
        values (Optional[List[float]]): A list of float values to initialize the matrix. Must represent the upper triangular part of the matrix. If None, initializes with zeros.

    Returns:
        _CorrMatrix: An instance of the `_CorrMatrix` data class.

    Raises:
        ValueError: If the number of values provided does not match the expected count for the upper triangular part of the matrix.

    Example:
        # Example of initializing a 3x3 correlation matrix
        matrix = new(size=3, values=[0.1, 0.2, 0.3])

    """
    import dataclasses
    import typing

    @dataclasses.dataclass
    class _CorrMatrix:
        size: int
        values: typing.List[float]

        def __post_init__(self):
            nv = len(self.values)
            nc = self.size * (self.size - 1) / 2
            if nv != nc:
                raise ValueError(
                    f"Invalid number of values {nv} for the correlation coefficients for the matrix size={self.size}. Should be {nc}."
                )

        def __repr__(self):
            return f"A_GIS.Math.Statistics.CorrelationMatrix.new(size={size},values={values})"

    return _CorrMatrix(size, values)



import pytest
import A_GIS.Image


def test_remove_metadata():
    """Test that metadata is removed."""

    im1 = A_GIS.Image.new(size=(10, 10), metadata={"hello": "world"})
    assert A_GIS.Image.Metadata.get(image=im1) != {}

    im2 = A_GIS.Image.Metadata.remove(image=im1)
    assert A_GIS.Image.Metadata.get(image=im2) == {}



def to_numpy(*, upper_tri):
    """Create a full correlation matrix from an upper triangular data set."""
    import numpy

    n = upper_tri.size
    matrix = numpy.eye(n)  # Start with an identity matrix
    upper_tri_indices = numpy.triu_indices(n, k=1)

    # Assign values to both upper and lower parts of the matrix
    matrix[upper_tri_indices] = upper_tri.values
    matrix[(upper_tri_indices[1], upper_tri_indices[0])] = (
        upper_tri.values  # Mirror the values
    )

    return matrix



def format(code: str) -> str:
    # Format with autopep8.
    import autopep8

    formatted_code = autopep8.fix_code(
        code, options={"aggressive": True, "aggressive": True}
    )

    # Format with docformatter.
    from docformatter import __main__ as main
    import io

    stdout = io.StringIO()
    stdin = io.StringIO(formatted_code)
    main._main(
        argv=["/path/to/docformatter", "--force-wrap", "-"],
        standard_out=stdout,
        standard_error=None,
        standard_in=stdin,
    )
    formatted_code = stdout.getvalue()

    # Format with black.
    import black

    BLACK_MODE = black.Mode(target_versions={black.TargetVersion.PY311}, line_length=79)
    try:
        formatted_code = black.format_file_contents(
            formatted_code, fast=False, mode=BLACK_MODE
        )
    except black.NothingChanged:
        pass
    finally:
        # Make sure there's a newline after the content
        if formatted_code and formatted_code[-1] != "\n":
            formatted_code += "\n"

    # Remove multiple new lines.
    import re

    formatted_code = re.sub(r"\n\s*\n", "\n\n", formatted_code)

    return formatted_code



def main():
    import A_GIS.Code.Tree
    import sys
    import pathlib

    tree = A_GIS.Code.Tree.recurse(path=pathlib.Path(sys.argv[1]))
    print(A_GIS.Code.Tree.print(tree=tree))

    A_GIS.Code.Tree.update(tree=tree)

    print(A_GIS.Code.Tree.print(tree=tree))


if __name__ == "__main__":
    main()



def sample(*, marginals, correlation, size=1):
    import scipy
    import numpy
    import A_GIS.Math.Statistics.CorrelationMatrix.to_numpy

    n = len(marginals)
    if n != correlation.size:
        raise ValueError(
            f"Incorrect size of correlation matrix {correlation.size} for the number of marginal distributions {n}."
        )

    # Create the full correlation correlation
    np_matrix = A_GIS.Math.Statistics.CorrelationMatrix.to_numpy(upper_tri=correlation)

    # Generate correlated normal random variables
    rvs = numpy.random.multivariate_normal(numpy.zeros(n), np_matrix, size=size)

    # Transform to uniform and then to triangular samples
    unifs = numpy.array([scipy.stats.norm.cdf(row) for row in rvs])
    realizations = numpy.array(
        [[marginals[i].ppf(u[i]) for i in range(n)] for u in unifs]
    )
    return realizations



import re


def sanitize_model_and_branch_names(model: str, branch: str = "main") -> (str, str):
    """
    Sanitizes and validates the model and branch names for downloading models.

    Args:
        model (str): The model name, potentially including URL and branch information.
        branch (str): The branch name with a default value of 'main'.

    Returns:
        (str, str): A tuple containing the sanitized model and branch names.

    Raises:
        ValueError: If the branch name contains invalid characters.

    Example:
        sanitized_model, sanitized_branch = sanitize_model_and_branch_names(
            model='https://huggingface.co/facebook/opt-1.3b', branch='dev'
        )
    """
    base = "https://huggingface.co"

    # Strip trailing slash from model name
    if model.endswith("/"):
        model = model[:-1]

    # Remove base URL if present
    if model.startswith(base + "/"):
        model = model[len(base) + 1 :]

    # Split model name and branch if included in model
    model_parts = model.split(":")
    model = model_parts[0] if len(model_parts) > 0 else model
    branch = model_parts[1] if len(model_parts) > 1 else branch

    # Validate branch name
    if branch is None:
        branch = "main"
    else:
        pattern = re.compile(r"^[a-zA-Z0-9._-]+$")
        if not pattern.match(branch):
            raise ValueError(
                "Invalid branch name. Only alphanumeric characters, period, underscore, and dash are allowed."
            )

    return model, branch



def from_numpy(*, matrix):
    """Extract the upper triangular values from a NumPy matrix and return a CorrelationMatrix object."""
    import A_GIS.Math.Statistics.CorrelationMatrix.new
    import numpy

    if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Input must be a square matrix.")

    n = matrix.shape[0]
    upper_tri_indices = numpy.triu_indices(n, k=1)

    # Extract the upper triangular values
    upper_tri_values = matrix[upper_tri_indices]

    # Convert to a list and create a CorrelationMatrix object
    return A_GIS.Math.Statistics.CorrelationMatrix.new(
        size=n, values=upper_tri_values.tolist()
    )



def start_download_threads(
    session, file_list, output_folder, start_from_scratch=False, threads=4
):
    import A_GIS.File.download
    from tqdm.contrib.concurrent import thread_map

    thread_map(
        lambda url: A_GIS.File.download(
            session, url, output_folder, start_from_scratch=start_from_scratch
        ),
        file_list,
        max_workers=threads,
        disable=True,
    )



def new(*, low: float, mode: float, high: float):
    """
    Return a triangular distribution class.

    """
    import dataclasses
    import typing

    @dataclasses.dataclass
    class _TriDist:
        low: float
        mode: float
        high: float

        def __post_init__(self):
            if low > mode:
                raise ValueError(
                    f"Minimum value low={low} cannot be greater than mode={mode}!"
                )
            if high < mode:
                raise ValueError(
                    f"Maximum value high={high} cannot be less than mode={mode}!"
                )
            self.c = (mode - low) / (high - low)

        def sample(self, size=1):
            import scipy.stats

            return scipy.stats.triang.rvs(
                self.c, loc=self.low, scale=self.high - self.low, size=size
            )

        def pdf(self, x):
            import scipy.stats

            return scipy.stats.triang.pdf(
                x, self.c, loc=self.low, scale=self.high - self.low
            )

        def cdf(self, x):
            import scipy.stats

            return scipy.stats.triang.cdf(
                x, self.c, loc=self.low, scale=self.high - self.low
            )

        def ppf(self, q):
            import scipy.stats

            return scipy.stats.triang.ppf(
                q, self.c, loc=self.low, scale=self.high - self.low
            )

        def __repr__(self):
            return f"A_GIS.Math.Statistics.Distribution.Triangular.new(low={self.low}, mode={self.mode}, high={self.high})"

    return _TriDist(low, mode, high)



import PIL.Image


def remove(*, image: PIL.Image.Image) -> PIL.Image.Image:
    """
    Generate a purified copy of a given image for consistent hashing.

    This function creates a 'pure' version of the input image by removing any additional metadata
    and retaining only the essential visual data. It's particularly useful for consistent hashing
    of image data, ensuring the hash is based solely on the image's visual content.

    Args:
        image (PIL.Image.Image): The image object to be purified. Must be a PIL Image object.
        custom_new (Optional[Callable[..., PIL.Image.Image]]): An optional custom function to create a new image object.
                                                              If None, `PIL.Image.new` is used as the default.

    Returns:
        PIL.Image.Image: A purified copy of the input image. This includes the same visual content
                        and color palette (if applicable), but without any additional metadata.

    Raises:
        ValueError: If `image` is not a PIL Image object.

    Note:
        The function ensures that the returned image is devoid of metadata like EXIF data,
        which can affect hash consistency.
    """

    if not isinstance(image, PIL.Image.Image):
        raise ValueError("Invalid image type. Expected a PIL Image object.")

    # Create a new image with the same mode and size as the original
    pure_image = PIL.Image.new(mode=image.mode, size=image.size)

    # Transfer pixel data from the original image to the new image
    pure_image.putdata(list(image.getdata()))

    # If the original image has a color palette, copy it to the new image
    if "P" in image.mode and hasattr(image, "getpalette"):
        pure_image.putpalette(image.getpalette())

    return pure_image



import pathlib


def recurse(
    *,
    path: pathlib.Path,
    ignore_list: set[str] = {"tests"},
    _root0: pathlib.Path = None,
) -> dict:
    """
    Generate a hierarchical dictionary of code structures within a directory.

    The function traverses a directory and its subdirectories to build a tree
    of the Python code present in each. It skips directories and files specified
    in the ignore list and processes Python files and packages.

    Args:
        path: A pathlib.Path object representing the directory to traverse.
        ignore_list: An optional set of strings representing directory or file names
                     to ignore during traversal.

    Returns:
        A dictionary representing the hierarchical structure of the code.
        Directories and Python files are keys, with their respective structures
        or content as values.

    """
    import A_GIS.File.read
    import A_GIS.Code.Tree.get

    # Quick return if we have asked for a file instead.
    if path.is_file():
        code = A_GIS.File.read(file=path)
        return A_GIS.Code.Tree.get(code=code)

    if _root0 is None:
        # Create entries list which is just the root directory/file.
        _root0 = path
        entries = [_root0]
    else:
        # If we are here from recursion, then we look at everything at this path.
        entries = list(path.iterdir())

    pkg_tree = {}
    for entry in entries:
        if entry.name in ignore_list or entry.name.startswith("__"):
            continue
        if entry.is_dir():
            pkg_file = pathlib.Path(entry) / "__init__.py"
            if pkg_file.exists():
                # Get the tree to test if a package.
                tree = A_GIS.Code.Tree.get(code=A_GIS.File.read(file=pkg_file))
                if tree == {}:
                    # It is a package.
                    tree[entry.name] = recurse(path=entry, _root0=_root0)
                    tree[entry.name]["_type"] = "package"
                # Add file annotations.
                for x in tree:
                    tree[x]["_file"] = str(pkg_file)
                pkg_tree.update(**tree)

    return pkg_tree



def get_patch(*, initial: str, final: str):
    """
    Generates the unified diff from the initial to final string.

    Parameters:
    initial: A string representing the initial state.
    final: A string representing the final state.

    Returns:
    A string representing the unified diff between the initial and final states.
    """
    from diff_match_patch import diff_match_patch

    dmp = diff_match_patch()
    patches = dmp.patch_make(initial, final)
    return dmp.patch_toText(patches)



import PIL.Image
import typing


def modify(
    *, image: PIL.Image.Image, metadata: typing.Optional[typing.Dict[str, str]] = None
) -> PIL.Image.Image:
    """
    Modify the metadata of an in-memory image object.

    This function updates the metadata of a given PIL.Image.Image object.
    The metadata is provided as a dictionary, with each key-value pair representing
    a metadata field and its value. The metadata values are serialized to JSON format
    before being added to the image.

    Args:
        image (PIL.Image.Image): The image object whose metadata is to be modified.
        metadata (typing.Optional[typing.Dict[str, str]]): A dictionary containing
            the metadata to be added or modified. Each value is serialized to JSON.
            Defaults to None.

    Returns:
        PIL.Image.Image: The modified image object with updated metadata.

    Examples:
        >>> import A_GIS.Image
        >>> image = A_GIS.Image.new(size=(100,100),metadata={'author1':'me'})
        >>> metadata = A_GIS.Image.Metadata.get(image=image)
        >>> metadata['author1']
        'me'
        >>> image = A_GIS.Image.Metadata.modify(image=image,metadata={'author2':'you'})
        >>> metadata = A_GIS.Image.Metadata.get(image=image)
        >>> metadata['author1']
        'me'
        >>> metadata['author2']
        'you'

    Raises:
        ValueError: If the provided metadata is not a dictionary.
    """
    import json
    import PIL.PngImagePlugin
    import A_GIS.File.Directory
    import pathlib

    if metadata is not None and not isinstance(metadata, dict):
        raise ValueError("Metadata must be a dictionary.")

    pnginfo = PIL.PngImagePlugin.PngInfo()
    # Copy existing meta data.
    for key, value in image.info.items():
        pnginfo.add_text(key, value)
    # Write/overwrite with new metadata.
    for key, value in metadata.items():
        pnginfo.add_text(key, json.dumps(value))
    tempdir = A_GIS.File.Directory.make(scoped_delete=True)
    path = pathlib.Path(tempdir.path) / "new.png"
    image.save(path, "PNG", pnginfo=pnginfo)
    return PIL.Image.open(path)



def download_model_files(
    session,
    model,
    branch,
    links,
    sha256,
    output_folder,
    progress_bar=None,
    start_from_scratch=False,
    threads=4,
    specific_file=None,
    is_llamacpp=False,
):
    import A_GIS.Data.Llm._helpers.start_download_threads
    import datetime
    import hashlib

    # Create the folder and writing the metadata
    output_folder.mkdir(parents=True, exist_ok=True)

    if not is_llamacpp:
        metadata = (
            f"url: https://huggingface.co/{model}\n"
            f"branch: {branch}\n"
            f'download date: {datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n'
        )

        sha256_str = "\n".join([f"    {item[1]} {item[0]}" for item in sha256])
        if sha256_str:
            metadata += f"sha256sum:\n{sha256_str}"

        metadata += "\n"
        (output_folder / "huggingface-metadata.txt").write_text(metadata)

    if specific_file:
        print(f"Downloading {specific_file} to {output_folder}")
    else:
        print(f"Downloading the model to {output_folder}")

    A_GIS.Data.Llm._helpers.start_download_threads(
        session,
        links,
        output_folder,
        start_from_scratch=start_from_scratch,
        threads=threads,
    )
